// base class for digraphs with effective test for isomorphism (O(n^log(n)))
// by Florian Kintzler & Torsten Langer, 1998

#include<iostream.h>
#include<stdlib.h>
#include<stdio.h>

#include "trigraph.h"
#include "trigraph.icc"


struct listeArray {
   int efrom;
   int eto;
   int order;
};
struct liste {
   int efrom;
   int eto;
   int order;
   liste* next;
};


// Die Klasse Graph speichert einen ungerichteten Graphen ohne multiple-edges.
class graph {
   public:
   graph(int n);
   virtual ~graph();
   int verbinde(int i,int j);
   int trenne(int i,int j);
   inline int verbunden(int i,int j) const;
   int deg(int i) const;
   inline int getKnotenzahl() const;
   // -------------- Isomorphietest -----------------
   // The algorithm of digraph isomorphism is based on the algorithm by
   // N. Deo, J. M. Davis and R. E. Lord
   // expected time complexity: O(n^log(n))
   int operator==(const graph& B);
   int operator!=(const graph& B);
   private:
   int* degArray() const;
   int wurzel(const int* adeg) const;
   int tiefensuche(liste* edgeliste,int knoten,int* namen) const;
   listeArray* tiefensuchbaum(int& m,int root,int* namen) const;
   int sortiere(listeArray* edge,int m) const;
   int isomorphie(const graph& B);
   int match(const graph& B,unsigned edgenr);  // the core of the test for
                                               // isomorphism
   listeArray* edge;  // contains the reordered edges
   unsigned* aPair;  // aPair and bPair contain an address table to reference
   unsigned* bPair;  // each other by isomorphic vertices. They can be made
                     // public if they're needed for special purpose.
   int m;  // number of edges in DFS-forest
   int* oldName;  // contains number of vertex; oldName = (namen)^-1
   // ------------Ende Isomorphietest ---------------
   char** adjazent;  // Adjazenzmatrix.
   unsigned knotenzahl;  // zur Speicherung der Knotenanzahl
};

// Konstruktor f"ur Graphen.
// Parameter: n=Knotenanzahl
// Der Konstruktor legt eine 2-dim Adjazenzmatrix an und initialisiert sie
// mit 0, d.h. es entsteht ein Graph mit n Knoten und 0 Kanten.
// Die Knoten werden dabei mit 1..n bezeichnet.
graph::graph(int n)
{
   int i;
   int j;
   // Speichere die Knotenanzahl.
   knotenzahl=n;
   // Lege Array von Arrays an. (n x n)-Matrix.
   adjazent = new char*[n];
   for(i=0;i<knotenzahl;i++){
      adjazent[i]=new char[i+1];
      for(j=0;j<=i;j++){
         adjazent[i][j]=0;
      }
   }
}

// Destruktor f"ur Graphen.
graph::~graph()
{
   int i;
   // Gebe Speicherbereich der Adjazenzmatrix wieder frei.
   for(i=0;i<knotenzahl;i++){
      delete[] adjazent[i];
   }
   delete[] adjazent;
}

// Gibt die Anzahl der Knoten des Graphen zur"uck.
inline int graph::getKnotenzahl() const
{
   return knotenzahl;
}

// Verbinde Knoten i mit Knoten j.
int graph::verbinde(int i,int j)
{
   int k = 0;
   // Liegen beide Knotenzahlen zwischen 1 und knotenzahl?
   if((i<1) || (i>getKnotenzahl()) || (j<1) || (j>getKnotenzahl())){
      return -1;
   }
   if(j>i) {
      k = i;
      i = j;
      j = k;
   }
   // Existiert bereits eine Kante zwischen i und j ?
   if(adjazent[i-1][j-1]==1){
      return 0;
   }
   // Verbinde i <-> j.
   adjazent[i-1][j-1]=1;
   return 1;
}

// L"ose die Verbindung i <-> j.
int graph::trenne(int i,int j)
{
   int k = 0;

   // Liegen beide Knotenzahlen zwischen 1 und knotenzahl ?
   if((i<1) || (i>getKnotenzahl()) || (j<1) || (j>getKnotenzahl())){
      return -1;
   }
   if(j>i) {
      k = i;
      i = j;
      j = k;
   }
   // Existiert keine Kante zwischen i und j ?
   if(adjazent[i-1][j-1]==0){
      return 0;
   }
   // L"ose Verbindung i -> j.
   adjazent[i-1][j-1]=0;
   return 1;
}

// Liefert zur"uck, ob eine Kante i <-> j besteht (1) oder nicht (0).
inline int graph::verbunden(int i,int j) const
{
   int k;

   if(j>i) {
      k = i;
      i = j;
      j = k;
   }

   return (int)adjazent[i-1][j-1];
}

// Gibt den Grad des Knoten i zur"uck.
int graph::deg(int i) const
{
   int sum=0;
   int j;

   if(i<1 || i>getKnotenzahl()) return -1;
   for(j=0;j<getKnotenzahl();j++){
      sum = sum + verbunden(i,j+1);
   }
   return sum;
}

// ---------------------------- Isomorphietest ------------------------------

// Die Funktion degArray() erstellt ein in aufsteigender Zahlenfolge
// sortiertes Array mit den deg()-Werten der zum Graphen geh"orenden Knoten.
int* graph::degArray() const
{
   int i;
   int j;
   int degree;
   int dummy;
   int* adeg = new int[getKnotenzahl()];
   // Erstelle sortiertes _DEG-Array.
   for(i=1;i<getKnotenzahl()+1;i++){
      degree = deg(i);
      // Sortiere aktuellen degree-wert in die Liste.
      for(j=1;j<i;j++){
         if(degree<adeg[j-1]) {
            dummy = adeg[j-1];
            adeg[j-1] = degree;
            degree = dummy;
         }
      }
      adeg[i-1]=degree;
   }
   return adeg;
}


int graph::wurzel(const int* adeg) const
{
   int i;
   int degree1;
   int degree2;
   int deg1vertex = 0;
   int counter1=1;
   int counter2=0;
   // Ermittle die kleinste Teilmenge mit gleichen deg-Werten.
   // z"ahle zun"achst die Menge der kleinsten deg-Nummern.
   degree1=adeg[0];
   for(i=1;adeg[i]==degree1 && i<getKnotenzahl();i++){
      counter1++;
   }
   // Nun vergleiche die ermittelte Zahl mit der Kardinalit"at der
   // nachfolgenden und vertausche, wenn gr"o"ser.
   while(i<getKnotenzahl())
   {
      degree2=adeg[i];
      deg1vertex = i;
      counter2=0;
      while(adeg[i]==degree2 && i<getKnotenzahl()){
         counter2++;
         i++;
      }
      if ( counter2<counter1 || degree1<(getKnotenzahl()+1)) {
         counter1=counter2;
         degree1=degree2;
      }
   }
   // Gib die Nummer des 1. Knotens mit Grad degree1 zur"uck.
   return deg1vertex+1;
}

// Berechnet rekursiv die einzelnen Tiefensuchwald-B"aume.
int graph::tiefensuche(liste* edgeliste,int knoten,int* namen) const
{
   int i;
   int j;
   int max;
   int twin = (0==1);
   liste* dummy;
   // F"ur jeden Knoten i ...
   for(i=1;i<getKnotenzahl()+1;i++){
      // .. kontrollieren, ob er mit knoten verbunden ist.
      if(verbunden(knoten,i)){
         // Ein Knoten ist gefunden.
         // Ist die verbindende Kante eine R"uckkante, also der gefundene
         // Knoten bereits besucht ?
         if(namen[i-1]!=0){
            // Ja, dann nehme die Kante einfach in die Liste auf.
            dummy=edgeliste;
            while(dummy->next!=NULL){
               if(dummy->next->efrom==namen[i-1]
               && dummy->next->eto==namen[knoten-1]) {
                  twin = (0==0);
                   // Kante ist schon in anderer Richtung aufgenommen
               }
               dummy=dummy->next;
            }
            if(!twin) {
               dummy->next = new liste;
               dummy=dummy->next;
               dummy->next=NULL;
               dummy->efrom=namen[knoten-1];
               dummy->eto=namen[i-1];
            }
            twin = (0==1);
         }
         else {
            // Nein, dann gebe dem neuen Knoten einen Namen, ...
            max = 0;
            for(j=0;j<getKnotenzahl();j++){
               if(namen[j]>max){
                  max=namen[j];
               }
            }
            namen[i-1] = max+1;
            // ... nehme ihn in die Liste auf, und ...
            dummy=edgeliste;
            while(dummy->next!=NULL){
               dummy=dummy->next;
            }
            dummy->next = new liste;
            dummy=dummy->next;
            dummy->next=NULL;
            dummy->efrom=namen[knoten-1];
            dummy->eto=namen[i-1];
            // ... erforsche weitere Wege von diesem Knoten aus:
            tiefensuche(edgeliste,i,namen);
         }
      }
   }
   return 1;
}


// Berechnet das Tiefensuchbaumarray (den Tiefensuchwald) , nach DDL.
listeArray* graph::tiefensuchbaum(int& m,int root,int* namen) const
{
   int i;
   int j;
   // Liste der m Kanten.
   liste* edgeliste = NULL;
   liste* dummy;
   // Array mit den neuen Namen anlegen. 0=noch nicht besucht.
   for(i=0;i<getKnotenzahl();i++){
      if(deg(i+1)==0){
         namen[i]=-1;
      }
      else {
         namen[i]=0;
      }
   }
   // Die erste Wurzel ist der 1. Knoten.
   namen[root-1]=1;
   do {
      if(edgeliste!=NULL) {
         // Eine Verbindung von der neuen "Generalwurzel" zu einem Suchbaum.
         dummy=edgeliste;
         while(dummy->next!=NULL){
           dummy=dummy->next;
         }
         dummy->next = new liste;
         dummy=dummy->next;
         dummy->next=NULL;
         dummy->efrom=0;
         dummy->eto=namen[root-1];
         dummy->order=0;
      }
      else {
         // 1. Verbindung von der neuen "Generalwurzel" zu einem Suchbaum.
         edgeliste = new liste;
         edgeliste->next=NULL;
         edgeliste->efrom=0;
         edgeliste->eto=1;
         edgeliste->order=0;
      }
      // Suche Baum von der Wurzel aus.
      tiefensuche(edgeliste,root,namen);
      // Ist noch eine Wurzel "uber, oder sind alle Knoten bereits besucht?
      root=-1;
      i=0;
      while(namen[i]!=0 && i<getKnotenzahl()){
         i++;
      }
      if(namen[i]==0 && i<getKnotenzahl()){
         root=i+1;
         i=0;
         for(j=0;j<getKnotenzahl();j++){
            if(namen[j]>i){
               i=namen[j];
            }
         }
         namen[root-1]=i+1;
      }
   } while (root!=-1);
   // Erzeuge ein Array aus den Kanten.
   m=0;
   dummy=edgeliste;
   while(dummy!=NULL){
      m++;
      dummy=dummy->next;
   }
   listeArray* array = new listeArray[m];
   dummy=edgeliste;
   for(i=0;i<m;i++){
      array[i].efrom=dummy->efrom;
      array[i].eto=dummy->eto;
      array[i].order=0;
      dummy=dummy->next;
      if(edgeliste != NULL){
         delete edgeliste;
         edgeliste = dummy;
      }
   }
   // Fertig.
   return array;
}

// Sortiert die gegebene Liste anhand der ORDER-Werte;
int graph::sortiere(listeArray* edge,int m) const
{
   int i;
   int j;
   int l;
   int dummy;
   for(i=2;i<m;i++){
      if(edge[i].efrom>edge[i].eto){
         l=edge[i].efrom;
      }
      else{
         l=edge[i].eto;
      }
      if(l<edge[i].order){
         for(j=i;j>1 && edge[j-1].order>=l;j--){
            dummy=edge[j].efrom;
            edge[j].efrom=edge[j-1].efrom;
            edge[j-1].efrom=dummy;
            dummy=edge[j].eto;
            edge[j].eto=edge[j-1].eto;
            edge[j-1].eto=dummy;
            edge[j].order=edge[j-1].order;
         }
      }
   }
   return 1;
}

int graph::match(const graph& B,unsigned edgenr)  // the edge "edgenr" is
{                                                 // being tested
   unsigned u = 0;  // edge "edgenr" = (u,v)
   unsigned v = 0;
   int y = 0;
   static int mc;  // counter to assure that the number of edges in B
                   // does not exceed the number of edges in A
   unsigned z = 0;  // two other vertices
   unsigned w = 0;
   int verify = (0==1);

   if(edgenr<=m) {
      if(edge[edgenr-1].efrom==0) u = 0;  // u,v from A
      else u = oldName[edge[edgenr-1].efrom-1];
      v = oldName[edge[edgenr-1].eto-1];
      if(u==0) {  // v is root
         for(y=1;y<=getKnotenzahl();y++) {  // y from B
            if(deg(v)==B.deg(y) && bPair[y-1]==0) {
               aPair[v-1] = y;
               bPair[y-1] = v;
               mc = 0;  // no edges in induced subgraph yet
               if(match(B,edgenr+1)) return (0==0);
               aPair[v-1] = 0;
               bPair[y-1] = 0;
            }
         }
      } else if(aPair[v-1]==0) {  // u is connected to a new vertex
         z = aPair[oldName[edge[edgenr-1].order-1]-1];  // check number of
         for(w=1;w<=getKnotenzahl();w++) {  // connections of z from B; w from B
            if(B.verbunden(z,w) && bPair[w-1]!=0) mc--;
         }
         if(mc!=0) return (0==1);  // B contains more edges than A
         for(y=1;y<=getKnotenzahl();y++) {  // y from B
            if(B.verbunden(aPair[u-1],y) && deg(v)==B.deg(y)
            && bPair[y-1]==0) {
               aPair[v-1] = y;
               bPair[y-1] = v;
               mc = 1;  // y is connected to one vertex (aPair[u-1]) now
               if(match(B,edgenr+1)) return (0==0);
               aPair[v-1] = 0;
               bPair[y-1] = 0;
            }
         }
      } else {  // u is connected to a vertex already matched
         y=1;  // y from A
         while(!verify && y<=getKnotenzahl()) {
            if(B.verbunden(aPair[u-1],y) && y==aPair[v-1]) verify = (0==0);
            y++;
         }
         if(verify) {
            mc++;
            if(match(B,edgenr+1)) return (0==0);
         }
      }
      return (0==1);
   } else return (0==0);
}

// Die Funktion isomorphie(graph& B) gibt 1 zur"uck, wenn die Graphen
// gleich sind, 0 sonst.
int graph::isomorphie(const graph& B)
{
   int i;
   int root;
   // Erstelle ADEG- und BDEG-Arrays.
   int* adeg = degArray();
   int* bdeg = B.degArray();
   // Sind Array ADEG und BDEG gleich?
   if(getKnotenzahl()!=B.getKnotenzahl()) return (0==1);
   if(getKnotenzahl()==0) return (0==0);
   for(i=0;i<getKnotenzahl();i++){
      if(adeg[i]!=bdeg[i]) {
         // Die Arrays sind nicht gleich: die Graphen sind nicht isomorph.
         return 0;
      }
   }
   // Hier anzukommen hei"st: die Graphen k"onnen isomorph sein.
   // Ermittle Wurzel
   root=wurzel(adeg);
   // Beginne den Tiefensuchbaum zu berechnen.
   // Array um die neuen Namen der Knoten zu Speichern.
   // aNamen[alterName]=neuerName.
   int* aNamen = new int[getKnotenzahl()];
   oldName = new int[getKnotenzahl()];
   edge=tiefensuchbaum(m,root,aNamen);
   // Einf"ugen der ORDER-Werte
   edge[0].order=0;
   for(i=1;i<m;i++){
      if(edge[i-1].efrom>edge[i-1].eto){
         edge[i].order=edge[i-1].efrom;
      }
      else {
         edge[i].order=edge[i-1].eto;
      }
      if(edge[i].order<edge[i-1].order) edge[i].order = edge[i-1].order;
      //each edge should have at least the order of the preceding edge
   }
   sortiere(edge,m);

   aPair = new unsigned[getKnotenzahl()];
   bPair = new unsigned[getKnotenzahl()];
   int result = 0;

   for(i=0;i<getKnotenzahl();i++) oldName[i] = -1;
   for(i=0;i<getKnotenzahl();i++) {  // create oldName = (aNamen)^-1
      if(aNamen[i]!=-1) oldName[aNamen[i]-1] = i+1;
   }
   for(i=0;i<getKnotenzahl();i++) {
      aPair[i] = 0;  // No matching vertices are known yet.
      bPair[i] = 0;
   }
   result = match(B,1);  // test for isomorphism is started with first edge

   delete[] aPair;
   delete[] bPair;
   delete[] oldName;
   delete[] edge;
   delete[] aNamen;
   delete[] adeg;
   delete[] bdeg;

   return result;
}

int graph::operator==(const graph& B)
{
   if(isomorphie(B)==1){
      return 1;
   }
   return 0;
}

int graph::operator!=(const graph& B)
{
   return !isomorphie(B);
}


// Der Isomorphietest fr trivalente Graphen
// Um den anderen Isomorphietest einfach zu benutzen, wird der Graph zuerst
// in die andere Graphenklasse kopiert und anschlieend der andere
// Isomorphietest aufgerufen.
bool IsomorphieTest(const TriGraph& t1, const TriGraph& t2)
{
   graph g1((int)t1.Knoten());
   graph g2((int)t2.Knoten());
   for(unsigned int i = 0;i < t1.Knoten();i++) {
      for(unsigned int j = i + 1;j < t1.Knoten();j++) {
         if(t1.Verbunden(i,j)) { g1.verbinde((int)i + 1,(int)j + 1); g1.verbinde((int)j + 1,(int)i + 1); }
         if(t2.Verbunden(i,j)) { g2.verbinde((int)i + 1,(int)j + 1); g2.verbinde((int)j + 1,(int)i + 1); }
      }
   }      
   return(g1 == g2);
}
